package be.iminds.ilabt.jfed.call_log_output;

import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingData;
import be.iminds.ilabt.jfed.util.library.DataConversionUtils;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import be.iminds.ilabt.jfed.util.library.XmlUtil;
import org.rendersnake.HtmlCanvas;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.regex.Pattern;

import static org.rendersnake.HtmlAttributesFactory.*;

/**
 * ApiTestResult: list of test method results
 */
public abstract class HtmlLogOutput extends LogOutput {
    private final String reportTitle;

    protected HtmlLogOutput(String reportTitle) {
        this.reportTitle = reportTitle;
    }

    public String getReportTitle() {
        return reportTitle;
    }

    public static String getEmbeddedImageCss(String cssClassName, String resourcename, int x, int y) throws IOException {
        InputStream is = HtmlLogOutput.class.getResourceAsStream(resourcename);
        assert is != null : "Failed to get InputStream for resource \""+resourcename+"\"";
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        byte[] buffer = new byte[1024];
        int len;
        while ((len = is.read(buffer)) != -1) {
            bout.write(buffer, 0, len);
        }
        bout.close();
        String base64 = DataConversionUtils.encodeBase64(bout.toByteArray());
        return "div.image-"+cssClassName+" {\n" +
                "  width:"+x+"px;\n" +
                "  height:"+y+"px;\n" +
                "  background-image:url(data:image/png;base64,"+base64+"); \n" +
                "}";
    }

    public static void writeImage(HtmlCanvas html, TestResultState state) throws IOException {
        String stateCssClassName = stateClass(state);

        //TODO: support non embedded images
        //  html.img(src(stateIconImage(state)));

        html.div(class_("image-"+stateCssClassName));
        html._div();
    }


    public static String stateClass(TestResultState state) {
        switch (state) {
            case SKIPPED:
                return "skip";
            case SUCCESS:
                return "success";
            case WARN:
                return "warn";
            case FAILED:
                return "fail";
            default:
                return "fail";
        }
    }

    public static void addKeyValue(HtmlCanvas html, String clazz, String key, String value) throws IOException {
        html.tr();
        html.td(class_("key"));
        html.span().content(key + ": ", true);
        html._td();
        html.td(class_("value"));
        html.span(class_(clazz)).content(value, true);
        html._td();
        html._tr();
    }

    public static void addCollapsable(HtmlCanvas html, String clazz, String shortText, String fullContent) throws IOException {
        if (fullContent == null) {
            addKeyValue(html, clazz, shortText, "");
        } else {
            int newLineCount = 0;
            for (int i = 0; i < fullContent.length(); i++)
                if (fullContent.charAt(i) == '\n')
                    newLineCount++;

            if (fullContent.length() > 150 || newLineCount > 1) {
                String id = clazz + Math.random();
                html.tr();
                html.td(class_("key"));
                html.span().content(shortText + ": ", true);
                html._td();
                html.td(class_("value"));
                html.a(class_("reveal").onClick("switchVisibility('" + id + "')").title("Hide/Show " + shortText)).content("Hide/Show", true);
                html.div(style("display:none;").class_(clazz + "-big").id(id)).content(fullContent, true);
                html._td();
                html._tr();
            } else {
                addKeyValue(html, clazz, shortText, fullContent);
            }
        }
    }

    public static void addClassVisibleToggle(HtmlCanvas html, String className, String text) throws IOException {
        html.div();
        html.span().content(text + ": ", true);
        html.a(class_("reveal").onClick("switchClassVisibility('" + className + "')").title("Hide/Show " + text)).content("Hide/Show", true);
        html._div();
    }

    public void toHtmlFile(File file) {
        toHtmlFile(file, false);
    }
    public abstract void htmlContent(HtmlCanvas html) throws IOException;

    public String toHtmlString(boolean showSecurityRisks) {
        return toHtmlString(showSecurityRisks, false);
    }
    public String toHtmlString(boolean showSecurityRisks, boolean addPolyFills) {
        StringWriter writer = null;
        try {
            writer = new StringWriter();
            toHtml(writer, showSecurityRisks, addPolyFills);
            return writer.getBuffer().toString();
        } catch (IOException e) {
            LOG.error("IOException writing to string", e);
            return null;
        } finally {
            if (writer != null)
                try {
                    writer.close();
                } catch (IOException e) {
                    LOG.error("IOException trying to close StringWriter", e);
                }
        }
    }
    public void toHtmlFile(File file, boolean showSecurityRisks) {
        FileWriter writer = null;
        try {
            writer = new FileWriter(file);
            toHtml(writer, showSecurityRisks);
        } catch (IOException e) {
            LOG.error("IOException writing to file \""+file.getPath()+"\"", e);
        } finally {
            if (writer != null)
                try {
                    writer.close();
                } catch (IOException e) {
                    LOG.error("IOException trying to close file \"" + file.getPath() + "\"", e);
                }
        }
    }
    public void toHtml(Writer writer, boolean showSecurityRisks) throws IOException {
        toHtml(writer, showSecurityRisks, false);
    }

    public static void writeStart(@Nonnull HtmlCanvas html, boolean addPolyFills, String reportTitle) throws IOException {
        String failCss = "";
        String okCss = "";
        String skipCss = "";
        String warnCss = "";

        try {

            failCss = getEmbeddedImageCss("fail", "error15x15.png", 15, 15);
            okCss = getEmbeddedImageCss("success", "ok15x15.png", 15, 15);
            skipCss = getEmbeddedImageCss("skip", "skip15x15.png", 15, 15);
            warnCss = getEmbeddedImageCss("warn", "warning15x15.png", 15, 15);
        } catch (IOException e) {
            LOG.error("Error while loading generated html css or png. This will be ignored.", e);
        }

        html
                .html()
                .head()
                .title().content(reportTitle)
                .meta(name("description").add("content", "description", false))
                .macros().stylesheet("result.css");

        //            if (embedCss) {
        InputStream cssis = HtmlLogOutput.class.getResourceAsStream("result.css");
        String cssContent = IOUtils.streamToString(cssis, "UTF-8");
        //                if (embedPng) {
        cssContent += failCss+"\n";
        cssContent += okCss+"\n";
        cssContent += skipCss+"\n";
        cssContent += warnCss+"\n";
        //                }
        html.style(media("screen").type("text/css")).content(cssContent, false);
        //            }

        //javascript
        InputStream jsis = HtmlLogOutput.class.getResourceAsStream("result.js");
        String jsContent = IOUtils.streamToString(jsis, "UTF-8");
        html.script(language("JavaScript1.3")).content(jsContent, false);

        jsis = HtmlLogOutput.class.getResourceAsStream("vkbeautify.0.99.00.beta.js");
        jsContent = IOUtils.streamToString(jsis, "UTF-8");
        html.script(language("JavaScript1.3")).content(jsContent, false);

        jsis = HtmlLogOutput.class.getResourceAsStream("base64js.min.js");
        jsContent = IOUtils.streamToString(jsis, "UTF-8");
        html.script(language("JavaScript1.3")).content(jsContent, false);

        jsis = HtmlLogOutput.class.getResourceAsStream("inflate.min.js");
        jsContent = IOUtils.streamToString(jsis, "UTF-8");
        html.script(language("JavaScript1.3")).content(jsContent, false);

        if (addPolyFills) {
            jsis = HtmlLogOutput.class.getResourceAsStream("encoding-indexes.js");
            jsContent = IOUtils.streamToString(jsis, "UTF-8");
            html.script(language("JavaScript1.3")).content(jsContent, false);

            jsis = HtmlLogOutput.class.getResourceAsStream("encoding.js");
            jsContent = IOUtils.streamToString(jsis, "UTF-8");
            html.script(language("JavaScript1.3")).content(jsContent, false);
        }

        html._head().body();

    }

    public static void writeEnd(@Nonnull HtmlCanvas html) throws IOException {
        html._body()._html();
    }

    public void toHtml(Writer writer, boolean showSecurityRisks, boolean addPolyFills) throws IOException {
        disableRemoveSecurityRisks = showSecurityRisks; //a bit of a hack now, would be nice to do this cleaner


        HtmlCanvas html = new HtmlCanvas(writer);

        writeStart(html, addPolyFills, reportTitle);

        htmlContent(html);

        writeEnd(html);

        writer.flush();

        disableRemoveSecurityRisks = false;
    }

    public static void writeOverviewItem(@Nonnull HtmlCanvas html, int index, TestResultState state, String name) throws IOException {
        html.div(class_("methodoverview-full"));
        writeImage(html, state);
        html.span(class_("methodoverview-methodname " + stateClass(state) + "header"))
                .span().a(href("#test" + index)).content(name, true)._span()
                ._span();
        html._div();
    }

    /*
        writeGroup(html, index++, res.state, res.methodName, res.description,
                   res.timeMs, new Date(res.startTimeMs), new Date(res.stopTimeMs)
                   res.exception, res.apiCallDetails, res.getLogLines())
    */

    public static void writeCall(HtmlCanvas html,
                                 final int index,
                                 final int apiCallCount,
                                 SerializableApiCallDetails details) throws IOException {
        String id = "api-call-" + index + "-" + apiCallCount;

        html.div();

        html.span().content("Api Call " + apiCallCount + ": ", true);
        html.a(class_("reveal").onClick("switchVisibility('" + id + "')").title("Hide/Show Api Call")).content("Hide/Show", true);
        html.span().content(" (" + details.getGeniMethodName() + " @ " + details.getAuthorityUrn() + ")");
        html._div();

        html.div(style("display:none;").class_("api-call").id(id));
        html.table(style("width:100%; table-layout: fixed;"));
        addKeyValue(html, "value-nondata", "API Name", details.getApiName());
        //                    html.div(class_("api")).content(details.getApiName(), true);
        if (details.getAuthorityName() != null || details.getAuthorityUrn() != null) {
            addKeyValue(html, "value-nondata", "Authority HRN", details.getAuthorityName());
            addKeyValue(html, "value-data", "Authority URN", details.getAuthorityUrn());
            //                        html.div(class_("authority-hrn")).content(details.getServer().getDefaultComponentManagerAsGeniUrn(), true);
            //                        html.div(class_("authority-urn")).content(details.getServer().getName(), true);
        }
        if (details.getStartTime() != null)
            addKeyValue(html, "value-data", "Start Time", dateToString(details.getStartTime()));
        if (details.getStopTime() != null)
            addKeyValue(html, "value-data", "Stop Time", dateToString(details.getStopTime()));
        addKeyValue(html, "value-data", "API Method Name", details.getGeniMethodName());
        addKeyValue(html, "value-data", "Java Method Name", details.getGeniMethodName());
        //                if (details.getMethodParameters() != null) {
        //                    String parametersString = "";
        //                    for (Map.Entry<String, String> e : details.getMethodParameters().entrySet()) {
        //                        parametersString += e.getKey()+":  "+e.getValue()+"\n";
        //                    }
        //                    addCollapsable(html, "value-data", "Parameters",
        //                            removeSecurityRisks(parametersString));
        //                }
        addKeyValue(html, "value-data", "Server URL", details.getCallServerUrl());
        if (!Objects.equals(details.getBaseServerUrl(), details.getCallServerUrl()))
            addKeyValue(html, "value-data", "Base Server URL", details.getBaseServerUrl());
        if (details.getConnectionSslAuthUserUrn() != null)
            addKeyValue(html, "value-data", "Connection User URN", details.getConnectionSslAuthUserUrn());
        if (details.getProxyInfo() != null)
            addKeyValue(html, "value-data", "Proxy", details.getProxyInfo().toString());
        if (details.getConnectionSslAuthUserCertificates() != null)
            addCollapsable(html, "value-data", "Connection User certificates", removeSecurityRisks(details.getConnectionSslAuthUserCertificates()));
        if (details.getProxyInfo() != null)
            addCollapsable(html, "value-data", "Connection Proxy Settings", removeSecurityRisks(details.getProxyInfo().toString()));
        //                    html.div(class_("api-method-name")).content(details.getGeniMethodName(), true);
        //                    html.div(class_("api-method-name-java")).content(details.getGeniMethodName(), true);
        //                    html.div(class_("api-method-parameters")).content(details.getMethodParameters()+"", true);
        //                    html.div(class_("api-method-server-url")).content(details.getServerUrl(), true);
        //                    html.div(class_("xmlrpc-request")).content(details.getXmlRpcRequest()+"", true);
        //                    html.div(class_("xmlrpc-reply")).content(details.getXmlRpcResponse()+"", true);
        //                    html.div(class_("http-request")).content(details.getHttpRequest(), true);
        //                    html.div(class_("http-reply")).content(details.getHttpResponse(), true);


        String baseId = "c" + index + "-" + apiCallCount;

        boolean hasRequestHttpRspec = isJsonStringRspec(details.getHttpRequest());
        boolean hasRequestGeniRspec = hasRspec(details.getXmlRpcRequestJsonObject(), true, false);

        boolean hasRequestRSpec = hasRequestHttpRspec || hasRequestGeniRspec;

        String idShowHtmlReqButton = "showHtmlReqButton" + baseId;
        String idShowXmlReqButton = "showXmlReqButton" + baseId;
        String idShowRspecReqButton = "showRSpecReqButton" + baseId;
        String idShowHtmlReqButtonPretty = "showHtmlReqButtonPretty" + baseId;
        String idShowRspecReqButtonPretty = "showRSpecReqButtonPretty" + baseId;

        String idShowHtmlReqText = "showHtmlReqText" + baseId;
        String idShowXmlReqText = "showXmlReqText" + baseId;
        String idShowRSpecReqText = "showRSpecReqText" + baseId;
        String idShowHtmlReqTextPretty = "showHtmlReqTextPretty" + baseId;
        String idShowRSpecReqTextPretty = "showRSpecReqTextPretty" + baseId;

        String idShowNoReqText = "showNoReqText" + baseId;

        String[] allRequestButtons = {
                idShowHtmlReqButton, idShowXmlReqButton,
                idShowRspecReqButton, idShowHtmlReqButtonPretty,
                idShowRspecReqButtonPretty
        };
        String[] allRequestText = {
                idShowHtmlReqText, idShowXmlReqText,
                idShowRSpecReqText, idShowHtmlReqTextPretty,
                idShowRSpecReqTextPretty
        };

        html.tr();
        html.td(class_("key"));


        html.table(class_("format-table"));
        html.tr();
        html.td(class_("format-table-key").title("Content of the HTTP Request")).content("HTTP");
        html.td(class_("format-table-raw"));
        html.div(class_("reveal").id(idShowHtmlReqButton)
                .onClick(
                        generateToggleButtonCode(idShowHtmlReqButton, allRequestButtons) +
                                generateRevealCode(idShowHtmlReqText, idShowNoReqText, allRequestText)
                ))
                .content("raw", true);
        html._td();
        html.td(class_("format-table-formatted"));
        html.div(class_("reveal").id(idShowHtmlReqButtonPretty)
                .onClick(
                        generateToggleButtonCode(idShowHtmlReqButtonPretty, allRequestButtons) +
                                generateRevealCode(idShowHtmlReqTextPretty, idShowNoReqText, allRequestText)+
                                "formatXmlDiv('"+idShowHtmlReqText+"','"+idShowHtmlReqTextPretty+"');"
                )).content("pretty", true);
        html._td();
        html._tr();


        html.tr();
        html.td(class_("format-table-key").title("List of XML-RPC Request Parameters, encoded as JSON for easier reading")).content("XML-RPC");
        html.td(class_("format-table-raw"));
        html.div(class_("reveal").id(idShowXmlReqButton)
                .onClick(generateToggleButtonCode(idShowXmlReqButton, allRequestButtons) +
                        generateRevealCode(idShowXmlReqText, idShowNoReqText, allRequestText)))
                .content("raw", true);
        html._td();
        html.td(class_("format-table-formatted"));
        //nothing here ever, XML-RPC is pre-formatted
        html._td();
        html._tr();

        if (hasRequestRSpec) {
            String rspecSourceId;
            if (hasRequestGeniRspec) {
                rspecSourceId = idShowXmlReqText;
            } else {
                rspecSourceId = idShowHtmlReqText;
            }

            String showRspecRawCode;
            String showRspecFormattedCode;
            if (hasRequestGeniRspec) {
                showRspecRawCode = "getEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReqText + "');\n";
                showRspecFormattedCode = "formatEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReqTextPretty + "');\n";
            } else {
                showRspecRawCode = "copyDiv('" + rspecSourceId + "', '" + idShowRSpecReqText + "');\n";
                showRspecFormattedCode = "formatXmlDiv('" + rspecSourceId + "', '" + idShowRSpecReqTextPretty + "');\n";
            }


            html.tr();
            html.td(class_("format-table-key").title("An RSpec detected somewhere in the Request.")).content("RSpec");
            html.td(class_("format-table-raw"));
            html.div(class_("reveal").id(idShowRspecReqButton)
                    .onClick(generateToggleButtonCode(idShowRspecReqButton, allRequestButtons) +
                            generateRevealCode(idShowRSpecReqText, idShowNoReqText, allRequestText)+
                            showRspecRawCode
                    )).content("raw", true);
            html._td();
            html.td(class_("format-table-formatted"));
            html.div(class_("reveal").id(idShowRspecReqButtonPretty)
                    .onClick(generateToggleButtonCode(idShowRspecReqButtonPretty, allRequestButtons) +
                            generateRevealCode(idShowRSpecReqTextPretty, idShowNoReqText, allRequestText)+
                            showRspecFormattedCode
                    )).content("pretty", true);
            html._td();
            html._tr();
        }

        html._table();

        html.div(class_("format-header"))
                .content( "Request: ", true);

        html._td();
        html.td(class_("value"));
        html.div(class_("value-data-big initially-hidden").id(idShowHtmlReqText)).content(
                removeSecurityRisks(details.getHttpRequest()), true);
        html.div(class_("value-data-big initially-hidden").id(idShowHtmlReqTextPretty)).content(
                "loading...", true);
        html.div(class_("value-data-big initially-hidden").id(idShowXmlReqText)).content(
                removeSecurityRisks(details.getXmlRpcRequestJsonString()), true);
        if (hasRequestRSpec) {
            html.div(class_("value-data-big initially-hidden")
                    .id(idShowRSpecReqText))
                    .content("loading...", true);

            html.div(class_("value-data-big initially-hidden")
                    .id(idShowRSpecReqTextPretty))
                    .content("loading...", true);
        }
        html.div(class_("small-help").id(idShowNoReqText)).content(
                "(click a format to show the data)", true);
        //                html._div();
        html._td();

        html._tr();


        String idShowHtmlReplyButton = "showHtmlReplyButton" + baseId;
        String idShowXmlReplyButton = "showXmlReplyButton" + baseId;
        String idShowGeniValReplyButton = "showGeniValReplyButton" + baseId;
        String idShowRSpecReplyButton = "showRSpecReplyButton" + baseId;

        String idShowHtmlReplyButtonFormatted = "showHtmlReplyButtonFormatted" + baseId;
        String idShowGeniValReplyButtonFormatted = "showGeniValReplyButtonFormatted" + baseId;
        String idShowRSpecReplyButtonFormatted = "showRSpecReplyButtonFormatted" + baseId;

        String idShowHtmlReplyText = "showHtmlReplyText" + baseId;
        String idShowXmlReplyText = "showXmlReplyText" + baseId;
        String idShowGeniValReplyText = "showGeniValReplyText" + baseId;
        String idShowGeniValReplyTextDetail = "showGeniValReplyTextDetail" + baseId;
        String idShowRSpecReplyText = "showRSpecReplyText" + baseId;

        String idShowHtmlReplyTextFormatted = "showHtmlReplyTextFormatted" + baseId;
        String idShowGeniValReplyTextFormatted = "showGeniValReplyTextFormatted" + baseId;
        String idShowRSpecReplyTextFormatted = "showRSpecReplyTextFormatted" + baseId;

        String idShowNoReplyText = "showNoReplyText" + baseId;

        String[] allReplyButtons = {
                idShowHtmlReplyButton, idShowXmlReplyButton,
                idShowGeniValReplyButton, idShowRSpecReplyButton,
                idShowHtmlReplyButtonFormatted,
                idShowGeniValReplyButtonFormatted, idShowRSpecReplyButtonFormatted,
        };
        String[] allReplyText = {
                idShowHtmlReplyText, idShowXmlReplyText,
                idShowGeniValReplyText, idShowRSpecReplyText,
                idShowHtmlReplyTextFormatted,
                idShowGeniValReplyTextFormatted, idShowRSpecReplyTextFormatted
        };

        boolean hasGeniVal = details.getGeniResponseCode() != null;

        Object xmlRpcResponseJsonObject = details.getXmlRpcResponseJsonObject();
        Object xmlRpcResponseJsonObjectValue;
        if (xmlRpcResponseJsonObject instanceof Map && ((Map)xmlRpcResponseJsonObject).containsKey("value")) {
            xmlRpcResponseJsonObjectValue = ((Map)xmlRpcResponseJsonObject).get("value");
        } else {
            xmlRpcResponseJsonObjectValue = xmlRpcResponseJsonObject;
        }

        boolean hasReplyHttpRspec = isRSpec(details.getHttpResponse());
        boolean hasReplyGeniRSpec = isRSpec(xmlRpcResponseJsonObjectValue);
        boolean hasReplyCompressedRSpec = isCompressedRSpec(xmlRpcResponseJsonObjectValue);
        boolean hasReplyRSpec = hasReplyCompressedRSpec || hasReplyHttpRspec || hasReplyGeniRSpec;

        boolean hasEmbeddedReplyGeniRSpec = !hasReplyRSpec && hasRspec(xmlRpcResponseJsonObjectValue, true, false);
        boolean hasEmbeddedReplyGeniCompressedRSpec = !hasReplyRSpec && hasRspec(xmlRpcResponseJsonObjectValue, false, true);
        boolean hasEmbeddedReplyRSpec = hasEmbeddedReplyGeniRSpec || hasEmbeddedReplyGeniCompressedRSpec;
        //TODO: we'll assume this is in "value" , "geni_rspec" !

        boolean hasFormattableGeniVal = hasGeniVal &&
                isXml(details.getXmlRpcGeniResponseValue()) &&
                !hasReplyCompressedRSpec;

        LOG.debug("call={} hasRspec={} hasCompressedRSpec={} hasGeniRSpec={} hasHttpRspec={} hasFormattableGeniVal={} resp.len={} hasComp={}",
                details.getGeniMethodName(),
                hasReplyRSpec, hasReplyCompressedRSpec, hasReplyGeniRSpec, hasReplyHttpRspec, hasFormattableGeniVal,
                (details.getXmlRpcGeniResponseValue() != null ? details.getXmlRpcGeniResponseValue().length()+"" : "null" ),
                hasCompressedReply(details.getXmlRpcRequestJsonString())
        );

        html.tr();
        html.td(class_("key"));


        html.table(class_("format-table"));
        html.tr();
        html.td(class_("format-table-key").title("Content of the HTTP Reply")).content("HTTP");
        html.td(class_("format-table-raw"));
        html.div(class_("reveal").id(idShowHtmlReplyButton)
                .onClick(
                        generateToggleButtonCode(idShowHtmlReplyButton, allReplyButtons) +
                                generateRevealCode(idShowHtmlReplyText, idShowNoReplyText, allReplyText)
                ))
                .content("raw", true);
        html._td();
        html.td(class_("format-table-formatted"));
        html.div(class_("reveal").id(idShowHtmlReplyButtonFormatted)
                .onClick(
                        generateToggleButtonCode(idShowHtmlReplyButtonFormatted, allReplyButtons) +
                                generateRevealCode(idShowHtmlReplyTextFormatted, idShowNoReplyText, allReplyText)+
                                "formatXmlDiv('"+idShowHtmlReplyText+"','"+idShowHtmlReplyTextFormatted+"');"
                )).content("pretty", true);
        html._td();
        html._tr();


        html.tr();
        html.td(class_("format-table-key").title("The XML-RPC reply, encoded as JSON for easier reading")).content("XML-RPC");
        html.td(class_("format-table-raw"));
        html.div(class_("reveal").id(idShowXmlReplyButton)
                .onClick(generateToggleButtonCode(idShowXmlReplyButton, allReplyButtons) +
                        generateRevealCode(idShowXmlReplyText, idShowNoReplyText, allReplyText)))
                .content("raw", true);
        html._td();
        html.td(class_("format-table-formatted"));
        //nothing here ever, XML-RPC is pre-formatted
        html._td();
        html._tr();

        if (hasGeniVal) {
            html.tr();
            html.td(class_("format-table-key").title("In the Geni SFA API's, all replies are a single object containing " +
                    "'code', 'output' and 'value' fields. These are shown seperately here. If the Geni value contains XML, " +
                    "\"pretty\" will format and ident that XML.")).content("Geni Value");
            html.td(class_("format-table-raw"));
            html.div(class_("reveal").id(idShowGeniValReplyButton)
                    .onClick(generateToggleButtonCode(idShowGeniValReplyButton, allReplyButtons) +
                            generateRevealCode(idShowGeniValReplyText, idShowNoReplyText, allReplyText)
                    )).content("raw", true);
            html._td();
            html.td(class_("format-table-formatted"));
            if (hasFormattableGeniVal) {
                html.div(class_("reveal").id(idShowGeniValReplyButtonFormatted)
                        .onClick(generateToggleButtonCode(idShowGeniValReplyButtonFormatted, allReplyButtons) +
                                generateRevealCode(idShowGeniValReplyTextFormatted, idShowNoReplyText, allReplyText)+
                                "formatXmlDiv('"+idShowGeniValReplyTextDetail+"','"+idShowGeniValReplyTextFormatted+"');\n"
                        )).content("pretty", true);
            }
            html._td();
            html._tr();
        }


        if (hasReplyRSpec || hasEmbeddedReplyRSpec) {
            String rspecSourceId;
            if (hasReplyCompressedRSpec || hasReplyGeniRSpec || hasEmbeddedReplyRSpec) {
                rspecSourceId = idShowGeniValReplyTextDetail;
            } else {
                rspecSourceId = idShowHtmlReplyText;
            }

            String showRspecRawCode;
            String showRspecFormattedCode;
            if (hasReplyCompressedRSpec) {
                showRspecRawCode = "decompressedDiv('" + rspecSourceId + "', '" + idShowRSpecReplyText + "');\n";
                showRspecFormattedCode = "formatCompressedXmlDiv('" + rspecSourceId + "', '" + idShowRSpecReplyTextFormatted + "');\n";
            } else if (hasReplyGeniRSpec) {
                showRspecRawCode = "getEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReplyText + "');\n";
                showRspecFormattedCode = "formatEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReplyTextFormatted + "');\n";
            } else if (hasReplyHttpRspec) {
                showRspecRawCode = "copyDiv('" + rspecSourceId + "', '" + idShowRSpecReplyText + "');\n";
                showRspecFormattedCode = "formatXmlDiv('" + rspecSourceId + "', '" + idShowRSpecReplyTextFormatted + "');\n";
            } else if (hasEmbeddedReplyGeniRSpec) {
                //we assume it's embedded in "geni_rspec"
                showRspecRawCode = "getEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReplyText + "');\n";
                showRspecFormattedCode = "formatEmbeddedRspecDiv('" + rspecSourceId + "', '" + idShowRSpecReplyTextFormatted + "');\n";
            } else if (hasEmbeddedReplyGeniCompressedRSpec) {
                //we assume it's embedded in "geni_rspec"
                showRspecRawCode = "decompressedDiv('" + rspecSourceId + "', '" + idShowRSpecReplyText + "');\n";
                showRspecFormattedCode = "formatCompressedXmlDiv('" + rspecSourceId + "', '" + idShowRSpecReplyTextFormatted + "');\n";
            } else {
                assert false : "unsupported case";
                showRspecRawCode = "error();";
                showRspecFormattedCode = "error();";
            }

            html.tr();
            html.td(class_("format-table-key").title("An RSpec found in the reply. " +
                    "If needed, base64 decoding and zlib inflation will be used to make the RSpec readable.")).content("RSpec");
            html.td(class_("format-table-raw"));
            html.div(class_("reveal").id(idShowRSpecReplyButton)
                    .onClick(generateToggleButtonCode(idShowRSpecReplyButton, allReplyButtons) +
                            generateRevealCode(idShowRSpecReplyText, idShowNoReplyText, allReplyText)+
                            showRspecRawCode
                    )).content("raw", true);
            html._td();
            html.td(class_("format-table-formatted"));
            html.div(class_("reveal").id(idShowRSpecReplyButtonFormatted)
                    .onClick(generateToggleButtonCode(idShowRSpecReplyButtonFormatted, allReplyButtons) +
                            generateRevealCode(idShowRSpecReplyTextFormatted, idShowNoReplyText, allReplyText)+
                            showRspecFormattedCode
                    )).content("pretty", true);
            html._td();
            html._tr();
        }

        html._table();

        html.div(class_("format-header"))
                .content( "Reply: ", true);


        html._td();
        html.td(class_("value"));
        html.div(class_("value-data-big initially-hidden").id(idShowHtmlReplyText)).content(
                removeSecurityRisks(details.getHttpResponse()), true);
        html.div(class_("value-data-big initially-hidden").id(idShowHtmlReplyTextFormatted)).content(
                "loading...", true);
        html.div(class_("value-data-big initially-hidden").id(idShowXmlReplyText)).content(
                removeSecurityRisks(details.getXmlRpcResponseJsonString()), true);
        if (hasGeniVal) {
            html.div(class_("value-data-big initially-hidden").id(idShowGeniValReplyTextFormatted)).content(
                    "loading...", true);
            html.div(class_("initially-hidden").id(idShowGeniValReplyText))
                    .div()
                    .span(class_("geni-header-key")).content("code: ")
                    .span(class_("value-data")).content(details.getGeniResponseCode() + "", false)
                    .span().content("&nbsp;", false)
                    .span(class_("value-data")).content(details.getGeniResponseCodeDescription(), true)
                    ._div()
                    .div()
                    .span(class_("geni-header-key")).content("output: ")
                    .div(class_("value-data-big")).content(details.getGeniResponseOutput(), true)
                    ._div()
                    .div()
                    .span(class_("geni-header-key")).content("value: ")
                    .div(class_("value-data-big")
                            .id(idShowGeniValReplyTextDetail))
                    .content(removeSecurityRisks(details.getXmlRpcGeniResponseValue()), true)
                    ._div()
                    ._div();
        }
        if (hasReplyRSpec || hasEmbeddedReplyRSpec) {
            html.div(class_("value-data-big initially-hidden")
                    .id(idShowRSpecReplyText))
                    .content("loading...", true);

            html.div(class_("value-data-big initially-hidden")
                    .id(idShowRSpecReplyTextFormatted))
                    .content("loading...", true);
        }
        html.div(class_("small-help").id(idShowNoReplyText)).content(
                "(click a format to show the data)", true);
        //                html._div();
        html._td();

        html._tr();

        String stitchingOverview = makeStitchingOverview(details);
        if (stitchingOverview != null)
            addCollapsable(html, "value-data", "Stitching Overview", stitchingOverview);

        if (details.getExceptionString() != null)
            addCollapsable(html, "value-data", "Exception", details.getExceptionString());
        html._table();
        html._div();
    }

    public static void writeGroup(HtmlCanvas html, int index,
                                  TestResultState state, String name, String description,
                                  long durationMs, Date start, Date stop,
                                  Throwable exception,
                                  List<SerializableApiCallDetails> calls,
                                  List<LogEntry> logEntries) throws IOException {
        writeGroupStart(html, index, state, name, description, durationMs, start, stop, exception, calls, logEntries);
        writeGroupEnd(html);
    }
    public static void writeGroup(HtmlCanvas html, int index,
                                  TestResultState state, String name, String description,
                                  long durationMs, Date start, Date stop,
                                  String exceptionStackTrace,
                                  List<SerializableApiCallDetails> calls,
                                  List<LogEntry> logEntries) throws IOException {
        writeGroupStart(html, index, state, name, description, durationMs, start, stop, exceptionStackTrace,
                calls, logEntries);
        writeGroupEnd(html);
    }
    public static void writeGroupStart(HtmlCanvas html, int index,
                                  TestResultState state, String name, String description,
                                  long durationMs, Date start, Date stop,
                                  Throwable exception,
                                  List<SerializableApiCallDetails> calls,
                                  List<LogEntry> logEntries) throws IOException {
        String exceptionStackTrace;
        if (exception != null) {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            exception.printStackTrace(pw);
            pw.close();
            exceptionStackTrace = removeSecurityRisks(sw.getBuffer().toString());
        } else {
            exceptionStackTrace = null;
        }

        writeGroupStart(html, index, state, name, description, durationMs, start, stop, exceptionStackTrace, calls, logEntries);
    }
    public static void writeGroupStart(HtmlCanvas html, int index,
                                       TestResultState state, String name, String description,
                                       long durationMs, Date start, Date stop,
                                       String exceptionStackTrace,
                                       List<SerializableApiCallDetails> calls,
                                       List<LogEntry> logEntries) throws IOException {
        html.span();
        html.div(class_("methoddetail-header " + stateClass(state) + "header"));//.content(res.methodName + "-> " +res.state, true);
        writeImage(html, state);
        html.span().a(id("test" + index)).content(name)._span();
        html._div();
        html.div(class_("methoddetail " + stateClass(state) + "body")); //.content(res.methodName + "-> " +res.state, true);
        html.div(class_("state")).content(state.toString());
        if (description != null)
            html.div(class_("description")).content(description);
        html.div(class_("times"));
        if (durationMs != 0) {
            html.span(class_("timeheader")).content("duration ");
            html.span(class_("value-nondata")).content((durationMs / 1000.0) + "s");
        }
        if (start != null) {
            html.span(class_("timeheader")).content(" from ");
            html.span(class_("value-nondata")).content(dateToString(start));
        }
        if (stop != null) {
            html.span(class_("timeheader")).content(" to ");
            html.span(class_("value-nondata")).content(dateToString(stop));
        }
        html._div();
        if (exceptionStackTrace != null) {
            addCollapsable(html, "value-data", "Exception Stacktrace", exceptionStackTrace);
        }
        int apiCallCount = 1;
        if (calls != null)
            for (SerializableApiCallDetails details : calls) {
                writeCall(html, index, apiCallCount++, details);
            }
        int logLineCount = 1;
        if (logEntries != null)
            for (LogEntry logline : logEntries) {
                String type = logline.type.name();

                //minor hack to change output meant for machines to debug level in this output meant for humans.
                if (logline.text.startsWith(MACHINE_LOGLINE_PREFIX)) {
                    type = LogLineType.DEBUG.name();
                }

                String logText = removeSecurityRisks(logline.text);

                String id = "logline-" + index + "-" + logLineCount;
                String className = "logline-" + type;
                String headerClassName = className + "-header";

                if (logline.getExceptionStackTrace() == null) {
                    html = addParagraphs(html.div(
                            logline.isFormatAsCode() ?
                                    class_(className + " format-as-code").id(id) :
                                    class_(className).id(id)
                    ).span(class_(headerClassName)).content(type + ": ", true), logText)._div();
                } else {
                    html = addParagraphs(html.div(
                            logline.isFormatAsCode() ?
                                    class_(className + " format-as-code").id(id) :
                                    class_(className).id(id)
                    ).span(class_(headerClassName)).content(type + ": ", true).div(), logText)._div();
                    String stacktrace = removeSecurityRisks(logline.getExceptionStackTrace());
                    html.div(class_("logline-exception"));
                    addCollapsable(html, "value-data", "Exception Stacktrace", stacktrace);
                    html._div()._div();
                }

                logLineCount++;
            }
    }
    public static void writeGroupEnd(HtmlCanvas html) throws IOException {
        html._div();
        html._span();
    }

    private static boolean isXml(String text) {
        return text != null && XmlUtil.isXmlLikeQuickTest(text.trim());
    }

    private static final Pattern HAS_COMPRESSED_PATTERN = Pattern.compile("\"?geni_compressed\"?\\s*:\\s*\"?true");
    public static boolean hasCompressedReply(String requestJsonString) {
        if (requestJsonString == null)
            return false;
        return HAS_COMPRESSED_PATTERN.matcher(requestJsonString).find();
    }

    private static String generateToggleButtonCode(String thisToggleButtonId, String... allToggleButtonIds) {
        Set<String> otherIds = new HashSet<>(Arrays.asList(allToggleButtonIds));
        otherIds.remove(thisToggleButtonId);

        String res = "var newVisible = switchRevealClassActive('" + thisToggleButtonId + "');\n";
        for (String otherId : otherIds) {
            res += "setRevealClassActive('" + otherId + "', false);\n";
        }
        return res;
    }

    public static String generateRevealCode(String toggledTextId, String noReplyTextId, String... allTextIds) {
        Set<String> otherIds = new HashSet<>(Arrays.asList(allTextIds));
        otherIds.remove(toggledTextId);
        otherIds.remove(noReplyTextId);

        String res = "setVisibility('"+toggledTextId+"', newVisible);\n"+
                "setVisibility('"+noReplyTextId+"', !newVisible);\n"+
                "if (newVisible) {\n";
        for (String otherId : otherIds) {
            res += "    setVisibility('" + otherId + "', false);\n";
        }
        res += "\n}\n";
        return res;
    }

    public static final String MACHINE_LOGLINE_PREFIX = "RAW_INFO";

    public static HtmlCanvas addParagraphs(HtmlCanvas html, String lines) {
        HtmlCanvas res = html;
        try {
            res = res.span();
            BufferedReader br = new BufferedReader(new StringReader(lines));
            String line = br.readLine();
            if (line != null) {
                res = res.write(line);
                line = br.readLine();
                while (line != null) {
                    res = res.div().content(line, true);
                    line = br.readLine();
                }
            }
            res = res._span();
        } catch (IOException e) {
            LOG.error("Did not expect an exception here", e);
        }
        return res;
    }


    public static String dateToString(long ms) {
        return dateToString(new Date(ms));
    }
    public static String dateToString(Date d) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS X");
        String iso8601LocalTZ = sdf.format(d);
        String rfc3339ZTZ = RFC3339Util.dateToRFC3339String(d, true/*forceZuluHack*/, true/* discardSubSeconds*/, true/*useZinsteadOfZero*/);
        return  iso8601LocalTZ + "(" +rfc3339ZTZ+ ")";
    }





    public static String makeStitchingOverview(String rspec) {

        if (rspec == null) {
//            return "Null RSpec";
            return null;
        }
        Element stitchingElement = StitchingData.extractStitchingElementFromRspec(rspec);
        if (stitchingElement == null) {
//            return("No stitching info in RSpec\n");
            return null;
        }

        List<Element> pathEls = quickDomChildrenHelper(stitchingElement, "path");
        if (pathEls == null || pathEls.isEmpty()) {
//            return("No paths in RSpec stitching info\n");
            return null;
        }
        String res = "";
        for (Element pathEl : pathEls) {
            String pathId = pathEl.getAttribute("id");
            assert pathId != null : "<stitching> has <path> element without id attribute: " + pathEl;

            res += ("  - path \"");
            res += (pathId);
            res += ("\"\n");

            NodeList hopList = pathEl.getElementsByTagName("hop");
            for (int i = 0; i < hopList.getLength(); i++) {
                org.w3c.dom.Node hopNode = hopList.item(i);
                assert hopNode != null;
                assert hopNode.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE;
                Element hopElement = (Element) hopNode;

                String hopId = hopElement.getAttribute("id");
                res += "    - hop id=\""+hopId+"\"\n";

                List<Element> linkEls = quickDomChildrenHelper(hopElement, "link");
                for (Element linkEl : linkEls) {
                    String linkIdUrn = linkEl.getAttribute("id");
                    res += ("      - link \"");
                    res += (linkIdUrn);
                    res += ("\"\n");

                    Element scDescriptorEl = quickDomChildHelper(linkEl, "switchingCapabilityDescriptor");
                    Element scSpecificInfoEl = quickDomChildHelper(scDescriptorEl, "switchingCapabilitySpecificInfo");
                    Element scSpecificInfo_L2scEl =
                            quickDomChildHelper(scSpecificInfoEl, "switchingCapabilitySpecificInfo_L2sc");

                    Element vlanRangeAvailabilityEl =
                            quickDomChildHelper(scSpecificInfo_L2scEl, "vlanRangeAvailability");
                    Element suggestedVLANRangeEl = quickDomChildHelper(scSpecificInfo_L2scEl, "suggestedVLANRange");

                    String vlanRangeAvailability = vlanRangeAvailabilityEl.getTextContent();
                    String suggestedVLANRange = suggestedVLANRangeEl.getTextContent();

                    res += ("        vlanRangeAvailability=\"");
                    res += (vlanRangeAvailability);
                    res += ("\"\n");

                    res += ("        suggestedVLANRange   =\"");
                    res += (suggestedVLANRange);
                    res += ("\"\n");
                }
            }
        }
        return res;
    }

    private static List<Element> quickDomChildrenHelper(Element e, String tagName) {
        NodeList resList = e.getElementsByTagName(tagName);
        List<Element> resElements = new ArrayList<>();
        for (int i = 0; i < resList.getLength(); i++) {
            org.w3c.dom.Node resNode = resList.item(i);
            assert resNode != null;
            assert resNode.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE;
            Element resElement = (Element) resNode;
            resElements.add(resElement);
        }
        return resElements;
    }

    private static Element quickDomChildHelper(Element e, String tagName) {
        NodeList resList = e.getElementsByTagName(tagName);
        assert resList.getLength() == 1;
        org.w3c.dom.Node resNode = resList.item(0);
        assert resNode != null;
        assert resNode.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE;
        return (Element) resNode;
    }


    private static boolean isJsonStringRspec(String jsonString) {
        //example (includes quotes!):    "<rspec...</rspec>"
        if (jsonString == null) return false;
        jsonString = jsonString.trim();
        if (!jsonString.startsWith("\"") && !jsonString.endsWith("\"")) return false;
        return isRSpec(jsonString.substring(0, jsonString.length()-1));
    }


    private static boolean isRSpec(@Nullable Object obj) {
        if (obj == null) return false;
        if (!(obj instanceof String)) return false;
        return isRSpec((String) obj);
    }
    private static boolean isRSpec(@Nullable String text) { //TODO: better test
        if (text == null || text.isEmpty()) return false;
        String head = text;
        if (text.length() > 100)
            head = text.substring(0, 100);
        return head.contains("<rspec");
    }

    private static boolean isCompressedRSpec(@Nullable Object obj) {
        if (obj == null) return false;
        if (!(obj instanceof String)) return false;
        return isCompressedRSpec((String) obj);
    }
    private static boolean isCompressedRSpec(@Nullable String text) {
        if (text == null || text.isEmpty()) return false;
        try {
            String decodedValue = DataConversionUtils.decompressFromBase64(text);
            return isRSpec(decodedValue);
        } catch (Exception ex) {
            return false;
        }
    }

    private static boolean hasRspec(@Nullable Object base, boolean matchNonCompressed, boolean matchCompressed) {
        if (base == null)
            return false;

        LinkedList<Object> objectsToCheck = new LinkedList<>();
        objectsToCheck.add(base);

        while (!objectsToCheck.isEmpty()) {
            Object o = objectsToCheck.removeLast(); //depth first
            //Object o = objectsToCheck.removeFirst(); //breadth first

            if (o instanceof Map) {
                Map h = (Map) o;
                Set<Map.Entry> entrySet = h.entrySet();
                for (Map.Entry e : entrySet) {
//                    Object key = e.getKey();
                    objectsToCheck.add(e.getValue());
                }
            } else if (o instanceof List) {
                for (Object val : (List) o)
                    objectsToCheck.add(val);
            } else if (o instanceof Object[]) {
                for (Object val : (Object[]) o)
                    objectsToCheck.add(val);
            } else if (o instanceof String) {
                String value = (String) o;
                if (matchNonCompressed && isRSpec(value)) {
                    return true;
                }
                if (matchCompressed && isCompressedRSpec(value)) {
                    return true;
                }
            } else {
                if (o instanceof Number || o instanceof Boolean) {
                    //do not log
                } else {
                    LOG.debug("Object is of type: " + o.getClass().getName());
                }
            }
        }
        return false;
    }

    private static List<String> extractRspecsFromJson(Object jsonRootObject) {
        if (jsonRootObject == null) return Collections.emptyList();

        List<String> res = new ArrayList<>();
        Deque<Object> objectsToProcess = new LinkedList<>();
        objectsToProcess.add(jsonRootObject);
        while (!objectsToProcess.isEmpty()) {
            Object jsonObject = objectsToProcess.pollFirst();
            if (jsonObject == null)
                continue;

            if (jsonObject instanceof String) {
                String s = (String) jsonObject;
                if (isRSpec(s))
                    res.add(s);
            }
            if (jsonObject instanceof List) {
                List v = (List) jsonObject;
                objectsToProcess.addAll(v);
            }
            if (jsonObject instanceof Map) {
                Map ht = (Map) jsonObject;
                objectsToProcess.addAll(ht.values());
            }
        }

        return res;
    }

    public static String makeStitchingOverview(SerializableApiCallDetails callDetails) {
        if (callDetails != null) {
            String stitchingOverviewText = "";
            int infoCount = 0;
            try {
                List<String> rspecsInRequest = extractRspecsFromJson(callDetails.getXmlRpcRequestJsonObject());

                if (!rspecsInRequest.isEmpty()) {
                    String requestRspecText = "Stitching info in request RSpec:\n";
                    int count = 0;
                    for (String rspec : rspecsInRequest) {
                        if (count > 0)
                            requestRspecText += "\n!!! There is an additional Rspec in the request:\n";
                        String rspecOverview = makeStitchingOverview(rspec);
                        if (rspecOverview != null) {
                            requestRspecText += rspecOverview;
                            count++;
                            infoCount++;
                        }
                    }
                    if (count > 0) {
                        stitchingOverviewText += requestRspecText+"\n\n";
                    }
                }
//                else
//                    stitchingOverviewText += "No stitching info parsed in request RSpec.\n";
                List<String> rspecsInReply = extractRspecsFromJson(callDetails.getXmlRpcResponseJsonObject());
                if (!rspecsInReply.isEmpty()) {
                    String replyRspecText = "Stitching info in reply RSpec:\n";
                    int count = 0;
                    for (String rspec : rspecsInReply) {
                        if (count > 0) replyRspecText += "\n!!! There is an additional Rspec in the reply:\n";
                        String rspecOverview = makeStitchingOverview(rspec);
                        if (rspecOverview != null) {
                            replyRspecText += rspecOverview;
                            count++;
                            infoCount++;
                        }
                    }
                    if (count > 0) {
                        stitchingOverviewText += replyRspecText+"\n\n";
                    }
                }
//                else
//                    stitchingOverviewText += ("No stitching info parsed in reply RSpec.\n");

                if (Objects.equals(callDetails.getGeniResponseCodeIsSuccess(), Boolean.FALSE)) { //note that this can be null
                    if (callDetails.getGeniResponseOutput().toLowerCase().contains("vlan") || callDetails.getGeniResponseOutput().toLowerCase().contains("any")) {
                        stitchingOverviewText += "Error reply with output \""+callDetails.getGeniResponseOutput()+"\""+"\n\n";
                        infoCount++;
                    }
                }
            } catch (AssertionError t) {
                LOG.error("AssertionError while processing stitching data: " + t.getMessage(), t);
                stitchingOverviewText += "\n\nException while processing stitching data: ";
                stitchingOverviewText += t.getMessage();
                infoCount++;
            } catch (Exception t) {
                LOG.error("Exception while processing stitching data: " + t.getMessage(), t);
                stitchingOverviewText += "\n\nException while processing stitching data: ";
                stitchingOverviewText += t.getMessage();
                infoCount++;
            }

            if (infoCount > 0)
                return stitchingOverviewText;
            else
                return null;
        } else
            return null;
    }
}
