package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.report.JobReport;
import be.iminds.ilabt.jfed.highlevel.jobs.report.JobReportWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.jobs.report.StitchingJobReport;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.RegisterSliversToSlaExperimentPartStateSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.jobs.states.ProvisionSliversOnAuthorityState;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 16:43
 */
public class ProvisionExperimentJob extends AbstractJob<Boolean> {
    private static final Logger ACTUAL_LOG = LoggerFactory.getLogger(ProvisionExperimentJob.class);

    private final JobStateFactory jobStateFactory;

    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final ExperimentPartControllerManager epcManager;

    private Boolean success = null;
    private final Logger LOG;

    public ProvisionExperimentJob(Experiment experiment, HighLevelTaskFactory hltf,
                                  TaskThread tt, JobStateFactory jobStateFactory,
                                  AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                  ExperimentPartControllerManager epcManager) {
        super("Provision Experiment " + experiment.getName(), experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        this.LOG = new JobReportWrappingLogger(ACTUAL_LOG, getJobReport());

        assert experiment.getRequestedEndTime() != null;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(experiment.getExperimentSpecification().getRspecs().get(0),
                        "assert failed: experiment.getExperimentSpecification() == null", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        LOG.debug("ProvisionExperimentJob started");

        if (experiment.getExperimentSpecification() != null &&
                experiment.getExperimentSpecificationFileManager() != null) {
            LOG.debug("ProvisionExperimentJob start background download of ExperimentSpecification files (when needed).");
            try {
                experiment.getExperimentSpecificationFileManager().fetchAll();
                if (experiment.getExperimentSpecificationFileManager().getErrorOccured())
                    throw new RuntimeException("Error fetching ESpec resources");
            } catch (Exception e) {
                LOG.debug("Failed to fetch some ESpec resources", e);
                updateMessage("Failed to fetch ESpec resources: "+e.getMessage());
                success = false;
                return false;
            }
        }

        final ProvisionPartsState provisionPartsState = new ProvisionPartsState();
        setAndRunState(provisionPartsState);

        if (provisionPartsState.isPermanentFailure()) {
            LOG.debug("ProvisionExperimentJob stopped in failure");

            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                        experiment.getExperimentSpecification().getRspecs().get(0),
                        "ProvisionExperimentJob stopped in failure", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        false, getExperimentClientIds());
            }

            success = false;
            return false;
        }

        //Now the Provision-step has passed without failure, we can register the slivers to the SLA-service
        if (experiment.isRequestSla()) {
            final RegisterSfaExperimentPartsToSlaCollectorState
                    registerSliversToSlaState = new RegisterSfaExperimentPartsToSlaCollectorState();
            setAndRunState(registerSliversToSlaState);
        }

        LOG.debug("ProvisionExperimentJob stopped successfully");

        success = true;
        return true;
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public class ProvisionPartsState extends SlicedState<ProvisionSfaExperimentPartJobSlice> {
        private final List<ProvisionSfaExperimentPartJobSlice> slices;

        protected ProvisionPartsState() {
            super("Provision all ExperimentParts");

            slices = experiment.getPartsStream()
                    .filter(part -> part.getState() != InternalState.FAILED)
                    .filter(part -> part instanceof SfaExperimentPart)
                    .map(part -> new ProvisionSfaExperimentPartJobSlice((SfaExperimentPart) part))
                    .collect(Collectors.toList());
        }


        public boolean isPermanentFailure() {
            for (ProvisionSfaExperimentPartJobSlice slice : slices) {
                if(slice.getResult() == ExperimentTaskStatus.FAILED)
                    return true;
            }
            return false;
        }

        @Override
        public Collection<ProvisionSfaExperimentPartJobSlice> getSlices() {
            return slices;
        }
    }

    public class ProvisionSfaExperimentPartJobSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public ProvisionSfaExperimentPartJobSlice(SfaExperimentPart experimentPart) {
            super(ProvisionExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsAllocateProvision(experimentPart.getConnectSfaAuthority())) {
                LOG.debug("Skipping provisioning on {}, as it does not support it.", experimentPart.getName());
                return ExperimentTaskStatus.SUCCESS;
            }

            ProvisionSliversOnAuthorityState provisionSliversOnAuthorityState =
                    jobStateFactory.createProvisionSliversOnAuthorityState(ProvisionExperimentJob.this, experimentPart, experiment.getUserSpecs());

            setAndRunState(provisionSliversOnAuthorityState);

            return provisionSliversOnAuthorityState.getStatus();
        }

        private boolean supportsAllocateProvision(Server auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.hasSeperateAllocateAndProvision();
        }
    }

    public class RegisterSfaExperimentPartsToSlaCollectorState extends SfaExperimentPartsSlicedState<RegisterSliversToSlaExperimentPartStateSlice> {

        private RegisterSfaExperimentPartsToSlaCollectorState() {
            super("Register resources at SLA Collector", getExperiment());
        }

        @Override
        public RegisterSliversToSlaExperimentPartStateSlice createJobSlice(SfaExperimentPart experimentPart) {
            return new RegisterSliversToSlaExperimentPartStateSlice(ProvisionExperimentJob.this, experimentPart, hltf);
        }
    }








    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
